# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Run ephemeral Docker Swarm services."""
from __future__ import annotations

from typing import TYPE_CHECKING

from docker import types

from airflow.exceptions import AirflowException
from airflow.providers.docker.operators.docker import DockerOperator
from airflow.utils.strings import get_random_string

if TYPE_CHECKING:
    from airflow.utils.context import Context


class DockerSwarmOperator(DockerOperator):
    """
    Execute a command as an ephemeral docker swarm service.

    Example use-case - Using Docker Swarm orchestration to make one-time
    scripts highly available.

    A temporary directory is created on the host and
    mounted into a container to allow storing files
    that together exceed the default disk size of 10GB in a container.
    The path to the mounted directory can be accessed
    via the environment variable ``AIRFLOW_TMP_DIR``.

    If a login to a private registry is required prior to pulling the image, a
    Docker connection needs to be configured in Airflow and the connection ID
    be provided with the parameter ``docker_conn_id``.

    :param image: Docker image from which to create the container.
        If image tag is omitted, "latest" will be used.
    :param api_version: Remote API version. Set to ``auto`` to automatically
        detect the server's version.
    :param auto_remove: Auto-removal of the container on daemon side when the
        container's process exits.
        The default is False.
    :param command: Command to be run in the container. (templated)
    :param docker_url: URL of the host running the docker daemon.
        Default is unix://var/run/docker.sock
    :param environment: Environment variables to set in the container. (templated)
    :param force_pull: Pull the docker image on every run. Default is False.
    :param mem_limit: Maximum amount of memory the container can use.
        Either a float value, which represents the limit in bytes,
        or a string like ``128m`` or ``1g``.
    :param tls_ca_cert: Path to a PEM-encoded certificate authority
        to secure the docker connection.
    :param tls_client_cert: Path to the PEM-encoded certificate
        used to authenticate docker client.
    :param tls_client_key: Path to the PEM-encoded key used to authenticate docker client.
    :param tls_hostname: Hostname to match against
        the docker server certificate or False to disable the check.
    :param tls_ssl_version: Version of SSL to use when communicating with docker daemon.
    :param tmp_dir: Mount point inside the container to
        a temporary directory created on the host by the operator.
        The path is also made available via the environment variable
        ``AIRFLOW_TMP_DIR`` inside the container.
    :param user: Default user inside the docker container.
    :param docker_conn_id: The :ref:`Docker connection id <howto/connection:docker>`
    :param tty: Allocate pseudo-TTY to the container of this service
        This needs to be set see logs of the Docker container / service.
    :param enable_logging: Show the application's logs in operator's logs.
        Supported only if the Docker engine is using json-file or journald logging drivers.
        The `tty` parameter should be set to use this with Python applications.
    :param configs: List of docker configs to be exposed to the containers of the swarm service.
        The configs are ConfigReference objects as per the docker api
        [https://docker-py.readthedocs.io/en/stable/services.html#docker.models.services.ServiceCollection.create]_
    :param secrets: List of docker secrets to be exposed to the containers of the swarm service.
        The secrets are SecretReference objects as per the docker create_service api.
        [https://docker-py.readthedocs.io/en/stable/services.html#docker.models.services.ServiceCollection.create]_
    :param mode: Indicate whether a service should be deployed as a replicated or global service,
        and associated parameters
    :param networks: List of network names or IDs or NetworkAttachmentConfig to attach the service to.
    :param placement: Placement instructions for the scheduler. If a list is passed instead,
        it is assumed to be a list of constraints as part of a Placement object.
    """

    def __init__(
        self,
        *,
        image: str,
        enable_logging: bool = True,
        configs: list[types.ConfigReference] | None = None,
        secrets: list[types.SecretReference] | None = None,
        mode: types.ServiceMode | None = None,
        networks: list[str | types.NetworkAttachmentConfig] | None = None,
        placement: types.Placement | list[types.Placement] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(image=image, **kwargs)
        self.enable_logging = enable_logging
        self.service = None
        self.configs = configs
        self.secrets = secrets
        self.mode = mode
        self.networks = networks
        self.placement = placement

    def execute(self, context: Context) -> None:
        self.environment["AIRFLOW_TMP_DIR"] = self.tmp_dir
        return self._run_service()

    def _run_service(self) -> None:
        self.log.info("Starting docker service from image %s", self.image)
        self.service = self.cli.create_service(
            types.TaskTemplate(
                container_spec=types.ContainerSpec(
                    image=self.image,
                    command=self.format_command(self.command),
                    mounts=self.mounts,
                    env=self.environment,
                    user=self.user,
                    tty=self.tty,
                    configs=self.configs,
                    secrets=self.secrets,
                ),
                restart_policy=types.RestartPolicy(condition="none"),
                resources=types.Resources(mem_limit=self.mem_limit),
                networks=self.networks,
                placement=self.placement,
            ),
            name=f"airflow-{get_random_string()}",
            labels={"name": f"airflow__{self.dag_id}__{self.task_id}"},
            mode=self.mode,
        )
        if self.service is None:
            raise Exception("Service should be set here")
        self.log.info("Service started: %s", str(self.service))

        # wait for the service to start the task
        while not self.cli.tasks(filters={"service": self.service["ID"]}):
            continue

        if self.enable_logging:
            self._stream_logs_to_output()

        while True:
            if self._has_service_terminated():
                self.log.info("Service status before exiting: %s", self._service_status())
                break

        self.log.info("auto_removeauto_removeauto_removeauto_removeauto_remove : %s", str(self.auto_remove))
        if self.service and self._service_status() != "complete":
            if self.auto_remove == "success":
                self.cli.remove_service(self.service["ID"])
            raise AirflowException("Service did not complete: " + repr(self.service))
        elif self.auto_remove == "success":
            if not self.service:
                raise Exception("The 'service' should be initialized before!")
            self.cli.remove_service(self.service["ID"])

    def _service_status(self) -> str | None:
        if not self.service:
            raise Exception("The 'service' should be initialized before!")
        return self.cli.tasks(filters={"service": self.service["ID"]})[0]["Status"]["State"]

    def _has_service_terminated(self) -> bool:
        status = self._service_status()
        return status in ["complete", "failed", "shutdown", "rejected", "orphaned", "remove"]

    def _stream_logs_to_output(self) -> None:
        if not self.service:
            raise Exception("The 'service' should be initialized before!")
        logs = self.cli.service_logs(
            self.service["ID"], follow=True, stdout=True, stderr=True, is_tty=self.tty
        )
        line = ""
        while True:
            try:
                log = next(logs)
            except StopIteration:
                # If the service log stream terminated, stop fetching logs further.
                break
            else:
                try:
                    log = log.decode()
                except UnicodeDecodeError:
                    continue
                if log == "\n":
                    self.log.info(line)
                    line = ""
                else:
                    line += log
        # flush any remaining log stream
        if line:
            self.log.info(line)

    def on_kill(self) -> None:
        if self.hook.client_created and self.service is not None:
            self.log.info("Removing docker service: %s", self.service["ID"])
            self.cli.remove_service(self.service["ID"])



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from functools import cached_property
from typing import TYPE_CHECKING, Any

from docker import APIClient, TLSConfig
from docker.constants import DEFAULT_TIMEOUT_SECONDS
from docker.errors import APIError

from airflow.exceptions import AirflowException, AirflowNotFoundException
from airflow.hooks.base import BaseHook

if TYPE_CHECKING:
    from airflow.models import Connection


class DockerHook(BaseHook):
    """
    Interact with a Docker Daemon and Container Registry.

    This class provide a thin wrapper around the ``docker.APIClient``.

    .. seealso::
        - :ref:`Docker Connection <howto/connection:docker>`
        - `Docker SDK: Low-level API <https://docker-py.readthedocs.io/en/stable/api.html?low-level-api>`_

    :param docker_conn_id: :ref:`Docker connection id <howto/connection:docker>` where stored credentials
         to Docker Registry. If set to ``None`` or empty then hook does not login to Container Registry.
    :param base_url: URL to the Docker server.
    :param version: The version of the API to use. Use ``auto`` or ``None`` for automatically detect
        the server's version.
    :param tls: Is connection required TLS, for enable pass ``True`` for use with default options,
        or pass a `docker.tls.TLSConfig` object to use custom configurations.
    :param timeout: Default timeout for API calls, in seconds.
    """

    conn_name_attr = "docker_conn_id"
    default_conn_name = "docker_default"
    conn_type = "docker"
    hook_name = "Docker"

    def __init__(
        self,
        docker_conn_id: str | None = default_conn_name,
        base_url: str | None = None,
        version: str | None = None,
        tls: TLSConfig | bool | None = None,
        timeout: int = DEFAULT_TIMEOUT_SECONDS,
    ) -> None:
        super().__init__()
        if not base_url:
            raise AirflowException("URL to the Docker server not provided.")
        elif tls:
            if base_url.startswith("tcp://"):
                base_url = base_url.replace("tcp://", "https://")
                self.log.debug("Change `base_url` schema from 'tcp://' to 'https://'.")
            if not base_url.startswith("https://"):
                self.log.warning("When `tls` specified then `base_url` expected 'https://' schema.")

        self.docker_conn_id = docker_conn_id
        self.__base_url = base_url
        self.__version = version
        self.__tls = tls or False
        self.__timeout = timeout
        self._client_created = False

    @staticmethod
    def construct_tls_config(
        ca_cert: str | None = None,
        client_cert: str | None = None,
        client_key: str | None = None,
        verify: bool = True,
        assert_hostname: str | bool | None = None,
        ssl_version: str | None = None,
    ) -> TLSConfig | bool:
        """
        Construct TLSConfig object from parts.

        :param ca_cert: Path to a PEM-encoded CA (Certificate Authority) certificate file.
        :param client_cert: Path to PEM-encoded certificate file.
        :param client_key: Path to PEM-encoded key file.
        :param verify: Set ``True`` to verify the validity of the provided certificate.
        :param assert_hostname: Hostname to match against the docker server certificate
            or ``False`` to disable the check.
        :param ssl_version: Version of SSL to use when communicating with docker daemon.
        """
        if ca_cert and client_cert and client_key:
            # Ignore type error on SSL version here.
            # It is deprecated and type annotation is wrong, and it should be string.
            return TLSConfig(
                ca_cert=ca_cert,
                client_cert=(client_cert, client_key),
                verify=verify,
                ssl_version=ssl_version,
                assert_hostname=assert_hostname,
            )
        return False

    @cached_property
    def api_client(self) -> APIClient:
        """Create connection to docker host and return ``docker.APIClient`` (cached)."""
        client = APIClient(
            base_url=self.__base_url, version=self.__version, tls=self.__tls, timeout=self.__timeout
        )
        if self.docker_conn_id:
            # Obtain connection and try to login to Container Registry only if ``docker_conn_id`` set.
            self.__login(client, self.get_connection(self.docker_conn_id))

        self._client_created = True
        return client

    @property
    def client_created(self) -> bool:
        """Is api_client created or not."""
        return self._client_created

    def get_conn(self) -> APIClient:
        """Create connection to docker host and return ``docker.APIClient`` (cached)."""
        return self.api_client

    def __login(self, client, conn: Connection) -> None:
        if not conn.host:
            raise AirflowNotFoundException("No Docker Registry URL provided.")
        if not conn.login:
            raise AirflowNotFoundException("No Docker Registry username provided.")

        registry = f"{conn.host}:{conn.port}" if conn.port else conn.host

        # Parse additional optional parameters
        email = conn.extra_dejson.get("email") or None
        reauth = conn.extra_dejson.get("reauth", True)
        if isinstance(reauth, str):
            reauth = reauth.lower()
            if reauth in ("y", "yes", "t", "true", "on", "1"):
                reauth = True
            elif reauth in ("n", "no", "f", "false", "off", "0"):
                reauth = False
            else:
                raise ValueError(f"Unable parse `reauth` value {reauth!r} to bool.")

        try:
            self.log.info("Login into Docker Registry: %s", registry)
            client.login(
                username=conn.login, password=conn.password, registry=registry, email=email, reauth=reauth
            )
            self.log.debug("Login successful")
        except APIError:
            self.log.error("Login failed")
            raise

    @staticmethod
    def get_connection_form_widgets() -> dict[str, Any]:
        """Returns connection form widgets."""
        from flask_appbuilder.fieldwidgets import BS3TextFieldWidget
        from flask_babel import lazy_gettext
        from wtforms import BooleanField, StringField

        return {
            "reauth": BooleanField(
                lazy_gettext("Reauthenticate"),
                description="Whether or not to refresh existing authentication on the Docker server.",
            ),
            "email": StringField(lazy_gettext("Email"), widget=BS3TextFieldWidget()),
        }

    @classmethod
    def get_ui_field_behaviour(cls) -> dict[str, Any]:
        """Returns custom field behaviour."""
        return {
            "hidden_fields": ["schema"],
            "relabeling": {
                "host": "Registry URL",
                "login": "Username",
            },
            "placeholders": {
                "extra": json.dumps(
                    {
                        "reauth": False,
                        "email": "Jane.Doe@example.org",
                    }
                )
            },
        }
